/* Common useful functions shared between objects */

import { TMRUtility } from "./tmr-utility.js";
import { RdDRollTables } from "./rdd-rolltables.js";

const level_category = { 
  "generale": "-4", 
  "particuliere": "-8", 
  "speciale": "-11", 
  "connaissance": "-11", 
  "draconic": "-11", 
  "melee": "-6", 
  "tir": "-8", 
  "lancer": "-8"
}
const competenceTroncs = [ ["Esquive", "Dague", "Corps à corps"],
                           ["Epée à 1 main", "Epée à 2 mains", "Hache à 1 main", "Hache à 2 mains", "Lance", "Masse à 1 main", "Masse à 2 mains"] ];
const competence_xp = {
  "-11" : [ 5, 10, 15, 25, 35, 45, 55, 70, 85, 100, 115, 135, 155, 175 ],
  "-8"  : [ 10, 20, 30, 40, 55, 70, 85, 100, 120, 140,160],
  "-6"  : [ 10, 20, 35, 50, 65, 80, 100, 120, 140],
  "-4"  : [ 15, 30, 45, 60, 80, 100, 120]
}

// This table starts at 0 -> niveau -10
const competence_xp_par_niveau = [ 5, 5, 5, 10, 10, 10, 10, 15, 15, 15, 15, 20, 20, 20, 20, 30, 30, 40, 40, 60, 60, 100, 100, 100, 100, 100, 100, 100, 100, 100];
const carac_array = [ "taille", "apparence", "constitution", "force", "agilite", "dexterite", "vue", "ouie", "odoratgout", "volonte", "intellect", "empathie", "reve", "chance", "melee", "tir", "lancer", "derobee"];
const difficultesLibres  = [0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10];
const ajustementsConditions  = [-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, +1, +2, +3, +4, +5, +6, +7, +8, +9, +10];
const fatigueMatrix = [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, // Dummy filler for the array.
                        [2, 3, 3, 2, 3, 3, 2, 3, 3, 2, 3, 3 ],
                        [2, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3 ],
                        [3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3 ],
                        [3, 3, 3, 3, 3, 4, 3, 3, 3, 3, 3, 4 ],
                        [3, 3, 4, 3, 3, 4, 3, 3, 4, 3, 3, 4 ],
                        [3, 3, 4, 3, 4, 4, 3, 3, 4, 3, 4, 4 ],
                        [3, 4, 4, 3, 4, 4, 3, 4, 4, 3, 4, 4 ],
                        [3, 4, 4, 4, 4, 4, 3, 4, 4, 4, 4, 4 ],
                        [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4 ],
                        [4, 4, 4, 4, 4, 5, 4, 4, 4, 4, 4, 5 ],
                        [4, 4, 5, 4, 4, 5, 4, 4, 5, 4, 4, 5 ],
                        [4, 4, 5, 4, 5, 5, 4, 4, 5, 4, 5, 5 ],
                        [4, 5, 5, 4, 5, 5, 4, 5, 5, 4, 5, 5 ],
                        [4, 5, 5, 5, 5, 5, 4, 5, 5, 5, 5, 5 ],
                        [5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5 ] ];
const fatigueMalus = [ 0, 0, 0, -1, -1, -1, -2, -3, -4, -5, -6, -7 ]; // Provides the malus for each segment of fatigue
const fatigueLineSize =  [ 3, 6, 7, 8, 9, 10, 11, 12];
const fatigueLineMalus = [ 0, -1, -2, -3, -4, -5, -6, -7 ];
const fatigueMarche = { "aise":      { "4":1, "6":2, "8":3, "10":4, "12":6 },
                        "malaise":   { "4":2, "6":3, "8":4, "10":6 },
                        "difficile": { "4":3, "6":4, "8":6 },
                        "tresdifficile": { "4":4, "6":6 } }
/* Static tables for commands /table */
const table2func = { "queues":  {descr: "queues : Tire une queue de Dragon", func: RdDRollTables.getQueue}, 
                     "ombre":   { descr: "ombre: Tire une Ombre de Dragon", func: RdDRollTables.getOmbre }, 
                     "tetehr":  {descr: "tetehr: Tire une Tête de Dragon pour Hauts Revants", fund: RdDRollTables.getTeteHR},
                     "tete"  :  { descr: "tete: Tire une Tête de Dragon", func: RdDRollTables.getTete},
                     "souffle": { descr: "souffle: Tire un Souffle de Dragon", func: RdDRollTables.getSouffle},
                     "tarot"  : { descr: "tarot: Tire une carte de Tarot Dracnique", func: RdDRollTables.getTarot} };

const definitionsEncaissement = {
  "mortel": [
    { minimum: undefined, maximum: 0, endurance: "0", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 1, maximum: 10, endurance: "1d4", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 11, maximum: 15, endurance: "1d6", vie: "0", legeres: 1, graves: 0, critiques: 0 },
    { minimum: 16, maximum: 19, endurance: "2d6", vie: "2", legeres: 0, graves: 1, critiques: 0 },
    { minimum: 20, maximum: undefined, endurance: "100", vie: "4 + @over20", legeres: 0, graves: 0, critiques: 1 },
  ],
  "non-mortel": [
    { minimum: undefined, maximum: 0, endurance: "0", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 1, maximum: 10, endurance: "1d4", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 11, maximum: 15, endurance: "1d6", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 16, maximum: 19, endurance: "2d6", vie: "0", legeres: 1, graves: 0, critiques: 0 },
    { minimum: 20, maximum: undefined, endurance: "100", vie: "1", legeres: 1, graves: 0, critiques: 0 },
  ],
  "cauchemar": [
    { minimum: undefined, maximum: 0, endurance: "0", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 1, maximum: 10, endurance: "1d4", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 11, maximum: 15, endurance: "1d6", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 16, maximum: 19, endurance: "2d6", vie: "0", legeres: 0, graves: 0, critiques: 0 },
    { minimum: 20, maximum: undefined, endurance: "3d6 + @over20", vie: "0", legeres: 0, graves: 0, critiques: 0 },
  ]
};

/* -------------------------------------------- */
export class RdDUtility  {
  
  /* -------------------------------------------- */
  static async preloadHandlebarsTemplates( ) {
    const templatePaths = [
      //Character Sheets
      'systems/foundryvtt-reve-de-dragon/templates/actor-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/actor-creature-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/actor-humanoide-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/actor-entite-sheet.html',
      //Items
      'systems/foundryvtt-reve-de-dragon/templates/item-competence-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-competencecreature-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-arme-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-armure-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-objet-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-conteneur-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-sort-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-herbe-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-ingredient-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-livre-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-tache-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-potion-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-rencontresTMR-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-queue-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-souffle-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-tarot-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-tete-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/item-ombre-sheet.html',
      'systems/foundryvtt-reve-de-dragon/templates/competence-categorie.html',
      'systems/foundryvtt-reve-de-dragon/templates/competence-carac-defaut.html',
      'systems/foundryvtt-reve-de-dragon/templates/competence-base.html',
      'systems/foundryvtt-reve-de-dragon/templates/enum-aspect-tarot.html',
      'systems/foundryvtt-reve-de-dragon/templates/enum-categorie.html',
      'systems/foundryvtt-reve-de-dragon/templates/enum-rarete.html',
      'systems/foundryvtt-reve-de-dragon/templates/arme-competence.html',
      'systems/foundryvtt-reve-de-dragon/templates/sort-draconic.html',
      'systems/foundryvtt-reve-de-dragon/templates/sort-tmr.html',
      // Dialogs
      'systems/foundryvtt-reve-de-dragon/templates/dialog-competence.html',
      'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-carac.html',
      'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-sort.html',
      'systems/foundryvtt-reve-de-dragon/templates/dialog-roll-encaisser.html',
      'systems/foundryvtt-reve-de-dragon/templates/dialog-tmr.html',
      // Calendrier
      'systems/foundryvtt-reve-de-dragon/templates/calendar-template.html',
      // Conteneur/item in Actor sheet
      'systems/foundryvtt-reve-de-dragon/templates/actor-inventaire-conteneur.html'
    ];

    return loadTemplates(templatePaths);
  }
  
  /* -------------------------------------------- */
  /** Construit la structure récursive des conteneurs, avec imbrication potentielle
   * 
   */
  static buildConteneur( objet, niveau ) {
    if (!niveau) niveau = 1;
    let str = Handlebars.partials['systems/foundryvtt-reve-de-dragon/templates/actor-inventaire-conteneur.html']( { item: objet} );
    if (objet.type == 'conteneur') {
      str = str + "<ol class='item-list'>";
      for (let subItem of objet.subItems) {
        str = str + this.buildConteneur(subItem, niveau+1);
      }
      str = str + "</ol>";
    }
    return new Handlebars.SafeString(str);
  }

  /* -------------------------------------------- */
  static buildResolutionTable( ) {
    let tableRes = []
    for (var j=0; j<=21; j++) {
      let subtab = [];
      for (var i=-10; i<=22; i++) {
        var m = (i + 10) * 0.5;
        var v;
        if (i == -9) {
          v = Math.floor(j / 2);
        } else if (i == -10) {
          v = Math.floor(j / 4);
        } else {
          if (j % 2 == 0) {
            var v = Math.ceil(j * m);
          } else {
            var v = Math.floor(j * m);
          }
        }
        if (v < 1) v = 1;
        let specResults
        if ( v > 100 )
          specResults = { part: Math.ceil(v / 5), epart: 1000, etotal: 1000 };
        else 
           specResults = specialResults[Math.ceil(v / 5 )];        
        let tabIndex = i+10;
        subtab[tabIndex] = { niveau: i, score: v, part: specResults.part, epart: specResults.epart, etotal: specResults.etotal } 
      }
      tableRes[j] = subtab;
    }
    return tableRes;
  }

  /* -------------------------------------------- */
  static getLevelCategory( )  
  {
    return level_category;
  }
  static getCaracArray()
  {
    return carac_array;
  }
  static getDifficultesLibres()
  {
    return difficultesLibres;
  }
  static getAjustementsConditions()
  {
    return ajustementsConditions;
  }

  /* -------------------------------------------- */
  static isTronc( compName )
  {
    for (let troncList of competenceTroncs) {
      for (let troncName of troncList) {
        if ( troncName == compName) 
          return troncList;
      }
    }
    return false;
  }

  /* -------------------------------------------- */
  static computeCompetenceXPCost( competence )
  {
    let minLevel = competence.data.base;
    if ( minLevel == competence.data.niveau) return 0;
    if ( competence.data.niveau < -10) return 0;

    let xp = 0;
    for (let i=minLevel+1; i<=competence.data.niveau; i++) {
       xp += competence_xp_par_niveau[i+10];
       //console.log(i, i+10, competence_xp_par_niveau[i+10]);
    }
    return xp;
  } 

  /* -------------------------------------------- */
  static computeCompetenceTroncXP( competenceList )
  {
    let xp = 0;
    for (let troncList of competenceTroncs) {
      let minNiveau = 15;
      for (let troncName of troncList) {
        let comp = RdDUtility.findCompetence( competenceList, troncName);
        minNiveau = (comp.data.niveau < minNiveau) ? comp.data.niveau : minNiveau; 
      }
      if ( minNiveau > 0 ) minNiveau = 0; // Clamp à 0, pour le tronc commun
      let minNiveauXP = competence_xp_par_niveau[minNiveau+10];
      xp += minNiveauXP;
      for (let troncName of troncList) {
        let comp = RdDUtility.findCompetence( competenceList, troncName);
        xp += competence_xp_par_niveau[comp.data.niveau+10] - minNiveauXP;
      }
    }
    return xp;
  }
  
  /* -------------------------------------------- */
  static computeCarac( data)
  {
    let fmax = parseInt(data.carac.taille.value) + 4;
    if ( data.carac.force.value > fmax ) 
        data.carac.force.value = fmax;
        
    data.carac.derobee.value = Math.floor(parseInt(((21 - data.carac.taille.value)) + parseInt(data.carac.agilite.value)) / 2);
    let bonusDomKey = Math.floor( (parseInt(data.carac.force.value) + parseInt(data.carac.taille.value)) / 2);
    data.attributs.plusdom.value = 2
    if (bonusDomKey < 8) 
      data.attributs.plusdom.value = -1;
    else if (bonusDomKey < 12) 
      data.attributs.plusdom.value = 0;
    else if (bonusDomKey < 14) 
      data.attributs.plusdom.value = 1;
      
    data.attributs.encombrement.value = (parseInt(data.carac.force.value) + parseInt(data.carac.taille.value)) / 2;
    data.carac.melee.value = Math.floor( (parseInt(data.carac.force.value) + parseInt(data.carac.agilite.value)) / 2);
    data.carac.tir.value = Math.floor( (parseInt(data.carac.vue.value) + parseInt(data.carac.dexterite.value)) / 2);
    data.carac.lancer.value = Math.floor( (parseInt(data.carac.tir.value) + parseInt(data.carac.force.value)) / 2);
    
    data.sante.vie.max = Math.ceil( (parseInt(data.carac.taille.value) + parseInt(data.carac.constitution.value)) /2 );
    if ( data.sante.vie.value > data.sante.vie.max)
      data.sante.vie.value = data.sante.vie.max;
    let endurance = Math.max( parseInt(data.carac.taille.value) + parseInt(data.carac.constitution.value), parseInt(data.sante.vie.max) + parseInt(data.carac.volonte.value) );
    data.sante.endurance.max = endurance;
    if ( data.sante.endurance.value > endurance)
      data.sante.endurance.value = endurance;
    data.sante.fatigue.max   = endurance*2;
    if ( data.sante.fatigue.value > data.sante.fatigue.max )
      data.sante.fatigue.value = data.sante.fatigue.max;
    
    data.attributs.sconst.value = 5; // Max !
    if ( data.carac.constitution.value < 9 ) 
      data.attributs.sconst.value = 2;
    else if (data.carac.constitution.value < 12 )
      data.attributs.sconst.value = 3;
    else if (data.carac.constitution.value < 15 )
      data.attributs.sconst.value = 4;      
    
    data.attributs.sust.value = 4; // Max !
    if ( data.carac.taille.value < 10 ) 
      data.attributs.sust.value = 2;
    else if (data.carac.taille.value < 14 )
      data.attributs.sust.value = 3;
      
    //Compteurs
    //data.compteurs.reve.value   = data.carac.reve.value;
    data.reve.reve.max = data.carac.reve.value;
    //data.compteurs.chance.value = data.carac.chance.value;
    data.compteurs.chance.max = data.carac.chance.value;
  }
  
  /* -------------------------------------------- */
  // Build the nice (?) html table used to manage fatigue.
  // max should Mbe the endurance max value
  static makeHTMLfatigueMatrix( value, max  ) 
  {
    max = (max < 16) ? 16 : max;
    max = (max > 30) ? 30 : max;
    value = (value > max*2) ? max*2 : value;
    value = (value < 0) ? 0 : value;
    
    let fatigueTab = fatigueMatrix[max];

    let table = $("<table/>").addClass('table-fatigue'); 
    let segmentIdx = 0;
    let fatigueCount = 0;
    for (var line=0; line < fatigueLineSize.length; line++) {
      let row = $("<tr/>");
      let segmentsPerLine = fatigueLineSize[line];
      row.append("<td class='fatigue-malus'>" + fatigueLineMalus[line] + "</td>");
      while (segmentIdx < segmentsPerLine) {
        let freeSize = fatigueTab[segmentIdx];
        for (let col=0; col <5; col++) {
          if ( col < freeSize ) {
           if (fatigueCount < value ) 
              row.append("<td class='fatigue-used'/>");
            else
              row.append("<td class='fatigue-free'/>");
            fatigueCount++;
          } else { 
            row.append("<td class='fatigue-none'/>");
          }
        }
        row.append("<td class='fatigue-separator'/>");
        segmentIdx = segmentIdx + 1;
      }
      table.append(row);
    }
    //console.log("fatigue", table);
    return table;
  }
  
  /* -------------------------------------------- */
  static getLocalisation( ) 
  {    
    let result = new Roll("d20").roll().total;
    let txt = ""
    if ( result <= 3 )  txt = "Jambe, genou, pied, jarret";
    else if ( result <= 7 )  txt = "Hanche, cuisse, fesse";
    else if ( result <= 9 )  txt = "Ventre, reins";
    else if ( result <= 12 ) txt = "Poitrine, dos";
    else if ( result <= 14 ) txt = "Avant-bras, main, coude";
    else if ( result <= 18 ) txt = "Epaule, bras, omoplate";
    else if ( result == 19)  txt = "Tête autre";
    else if ( result == 20)  txt = "Tête visage";
    
    return { result: result, label: txt };
  }

  /* -------------------------------------------- */
  static computeBlessuresSante( degats, mortalite="mortel" ) {
    let encaissement = RdDUtility.selectEncaissement(degats, mortalite)
    let over20 =  degats > 20 ? degats - 20 : 0 
    encaissement.endurance = - RdDUtility._evaluatePerte(encaissement.endurance, over20)
    encaissement.vie = - RdDUtility._evaluatePerte(encaissement.vie, degats, over20)
    return encaissement;
  }

  /* -------------------------------------------- */
  static selectEncaissement( degats, mortalite ) {
    const table = definitionsEncaissement[mortalite] === undefined ? definitionsEncaissement["mortel"] : definitionsEncaissement[mortalite];
    for (let encaissement of table) {
      if ((encaissement.minimum === undefined || encaissement.minimum <= degats)
        && (encaissement.maximum === undefined || degats <= encaissement.maximum)) {
        return duplicate(encaissement);
      }
    }
    return duplicate(table[0]);
  }


  /* -------------------------------------------- */
  static _evaluatePerte(formula, over20) {
    console.log("_evaluatePerte", formula, over20 )
    let perte = new Roll(formula, { over20:over20})
    perte.evaluate()
    return perte.total
  }
  
  /* -------------------------------------------- */
  static currentFatigueMalus( value, max)
  {
    max = (max < 16) ? 16 : max;
    max = (max > 30) ? 30 : max;
    value = (value > max*2) ? max*2 : value;
    value = (value < 0) ? 0 : value;
    
    let fatigueTab = fatigueMatrix[max];
    let fatigueRem = value;
    for (let idx=0; idx<fatigueTab.length; idx++) {
      fatigueRem -= fatigueTab[idx];
      if ( fatigueRem <= 0) {
        return fatigueMalus[idx];
      }
    }
    return -7; // This is the max !
  }
  
  /* -------------------------------------------- */  
  static findCompetence(compList, compName)  
  {
    compName = compName.toLowerCase();
    return compList.find(item => item.name.toLowerCase() == compName && (item.type =="competence" || item.type == "competencecreature"))
  }
  
  /* -------------------------------------------- */
  static getArmeCategory( arme )
  {
    let compname = arme.data.competence.toLowerCase(); 
    if ( compname.match("hache") ) return "hache";
    if ( compname.match("hast") ) return "hast";
    if ( compname.match("lance") ) return "lance";
    if ( compname.match("bouclier") ) return "bouclier";
    if ( compname.match("masse") ) return "masse";
    if ( compname.match("fléau") ) return "fleau";
    if ( compname.match("epée") ) { 
      let armename = arme.name.toLowerCase();
      if (armename == "dague" || armename.match("gnome") )
        return "epee_courte";
    }
    return "epee_longue";
  }
  
  /* -------------------------------------------- */
  static isArmeMelee( compName) 
  {
    let comp = compName.toLowerCase();
    if (comp.match("epée") || comp.match("épée") || comp.match("hache") || comp.match("fleau") || comp.match("mass") || comp.match("lance") || comp.match("hast") || comp == "dague" || comp=="bouclier")
      return true;
    return false;
  }
  
  /* -------------------------------------------- */
  static buildDefenseChatCard( attacker, target, rollData )
  {
    console.log("Attacker.defense", attacker, target, attacker.data._id, rollData.competence.data.categorie );
    let myTarget = target.actor;
    let defenseMsg = { title: "Défense en combat", 
                       content: "<strong>"+myTarget.name+"</strong> doit se défendre : <br><span class='chat-card-button-area'>" +
                                "<a class='chat-card-button' id='encaisser-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "'>Encaisser !</a></span>",                                                         
                       whisper: ChatMessage.getWhisperRecipients( myTarget.name ),
                       defenderid: myTarget.data._id,
                       rollMode: true
                      };
    
    if ( rollData.competence.data.categorie == 'melee' || rollData.competence.data.categorie == 'creature') { // Melee attack or creature
      let defenderArmes = [];
      for (const arme of myTarget.data.items) {
        if (arme.type == "arme" && this.isArmeMelee(arme.data.competence)) {
          defenderArmes.push( arme );
          defenseMsg.content +=  "<br><a class='chat-card-button' id='parer-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "' data-armeid='"+arme._id+"'>Parer avec " + arme.name + "</a></span>";
        }
      }
      defenseMsg.content +=  "<br><a class='chat-card-button' id='esquiver-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "'>Esquiver</a></span>";
    }
    if ( rollData.competence.data.categorie == "tir" ) {
      for (const arme of myTarget.data.items) { // Bouclier for parry
        if ( arme.type == "arme" && arme.name.toLowerCase.match("bouclier") ) {
          defenderArmes.push( arme );
          defenseMsg.content +=  "<br><a class='chat-card-button' id='parer-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "' data-armeid='"+arme._id+"'>Parer avec " + arme.name + "</a></span>";
        }
      }
    }
    if ( rollData.competence.data.categorie == "lancer" ) {
      for (const arme of myTarget.data.items) { // Bouclier for parry  Dodge/Esquive
        if ( arme.type == "arme" && arme.name.toLowerCase.match("bouclier") ) {
          defenderArmes.push( arme );
          defenseMsg.content +=  "<br><a class='chat-card-button' id='parer-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "' data-armeid='"+arme._id+"'>Parer avec " + arme.name + "</a></span>";
        }
      }
      defenseMsg.content +=  "<br><a class='chat-card-button' id='esquiver-button' data-attackerid='"+attacker.data._id + "' data-defenderid='" + myTarget.data._id + "'>Esquiver</a></span>";
    }
    
    defenseMsg.toSocket = true; // True per default for all players
    if (game.user.isGM) { // In GM case, only if target is a player
      defenseMsg.toSocket = myTarget.hasPlayerOwner;
    }

    return defenseMsg;
  }

  /* -------------------------------------------- */
  static performSocketMesssage( sockmsg ) 
  {
    console.log(">>>>> MSG RECV", sockmsg);
    if ( sockmsg.msg == "msg_encaisser" ) {
      if ( game.user.isGM ) { // Seul le GM effectue l'encaissement sur la fiche
        let rollData = game.system.rdd.rollDataHandler[sockmsg.data.attackerid]; // Retrieve the rolldata from the store
        let defenderActor = game.actors.get( sockmsg.data.defenderid );
        defenderActor.encaisserDommages( rollData );
      }
    } else if (sockmsg.msg == "msg_defense" ) {
      let defenderActor = game.actors.get( sockmsg.data.defenderid );
      if ( defenderActor ) {
        if ( (game.user.isGM && !defenderActor.hasPlayerOwner) || (defenderActor.hasPlayerOwner && (game.user.character.id == defenderActor.id) ) ) {
          console.log("User is pushing message...", game.user.name);
          defenderActor.currentRollData = duplicate(sockmsg.data.rollData);
          sockmsg.data.whisper = [ game.user ];
          sockmsg.data.blind   = true;
          sockmsg.data.rollMode = "blindroll";
          ChatMessage.create( sockmsg.data );
        }
      }
    }
  }
  
  /* -------------------------------------------- */
  static async chatListeners( html )
  {
    html.on("click", '#encaisser-button', event => {
      event.preventDefault();
      let attackerid = event.currentTarget.attributes['data-attackerid'].value;
      let defenderid = event.currentTarget.attributes['data-defenderid'].value;
      if ( game.user.isGM ) { // Current user is the GM -> direct access
        let rollData = game.system.rdd.rollDataHandler[attackerid];
        rollData.attackerid = attackerid;
        rollData.defenderid = defenderid;
        let defenderActor = game.actors.get(defenderid );
        defenderActor.encaisserDommages( rollData );
      } else { // Emit message for GM
        game.socket.emit("system.foundryvtt-reve-de-dragon", {
              msg: "msg_encaisser",
              data: { attackerid: attackerid, defenderid: defenderid  } 
          } );
      }
    });
    
    html.on("click", '#parer-button', event => {
      event.preventDefault();
      let defenderActor = game.actors.get(event.currentTarget.attributes['data-defenderid'].value );
      let armeId = event.currentTarget.attributes['data-armeid'].value;
      let rollData = defenderActor.currentRollData;
      defenderActor.parerAttaque( rollData, armeId );
    }); 

    html.on("click", '#esquiver-button', event => {
      event.preventDefault();
      let defenderActor = game.actors.get(event.currentTarget.attributes['data-defenderid'].value );
      let rollData = defenderActor.currentRollData;
      //console.log("Esquive !", rollData, defenderActor);
      defenderActor.esquiverAttaque( rollData );
    }); 

    html.on("click", '#particuliere-attaque', event => {
      event.preventDefault();
      let attackerActor = game.actors.get(event.currentTarget.attributes['data-attackerid'].value );
      let rollData = attackerActor.currentRollData;
      rollData.particuliereAttaque = game.actors.get(event.currentTarget.attributes['data-mode'].value );
      //console.log("Particulère !", rollData);
      attackerActor.continueRoll( rollData );
    }); 
     
  }

  /* -------------------------------------------- */
  /* Display help for /table */
  static displayHelpTable( msg )
  {
    msg.content = "";
    for (let [name, tableData] of Object.entries(table2func)) {
      msg.content += "<br>" + tableData.descr;
    }
    ChatMessage.create( msg );
  }

  /* -------------------------------------------- */
  /* Manage chat commands */
  static  processChatCommand( commands, content, msg ) {    
    // Setup new message's visibility
    let rollMode = game.settings.get("core", "rollMode");
    if (["gmroll", "blindroll"].includes(rollMode)) msg["whisper"] = ChatMessage.getWhisperRecipients("GM");
    if (rollMode === "blindroll") msg["blind"] = true;
    msg["type"] = 0;

    let command = commands[0];

    // Roll on a table
    if (command === "/table") {
      if ( commands[1] ) {
        let tableName = commands[1].toLowerCase();
        table2func[tableName].func();
      } else { 
        this.displayHelpTable( msg );
      }
      return false
    } else if (command === "/tmrr") {
        TMRUtility.getRencontre(commands[1], commands[2] )
        return false
    } else if (command === "/tmra") {
        TMRUtility.getTMRAleatoire( )
        return false
    }

    return true;
  }
}
